/*
 * Created on Nov 16, 2004
 * 
 */

/**
 * @author dengel
 *
 */

import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Random;

import jj2000.j2k.util.ArrayUtil;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.util.param.ParamKeyUtil;
import jj2000.j2k.wavelet.Subband;
import jj2000.j2k.wavelet.SubbandStructure;
import jj2000.j2k.wavelet.analysis.*;
import jj2000.j2k.wavelet.synthesis.SynWTParamFilterFloatLift9x7;
import jj2000.j2k.image.*;
import jj2000.j2k.image.edgedectection.Sobel;
import jj2000.j2k.image.input.*;


public class ParamTest {


	   /**
     * Prints the error message 'msg' to standard err, prepending "ERROR" to
     * it, and sets the exitCode to 'code'. An exit code different than 0
     * indicates that there where problems.
     *
     * @param msg The error message
     *
     * @param code The exit code to set
     * */
    private static void error(String msg) {
        FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,msg);
    }
    
    public static void printFilterCoefficients(float alpha)
    {
    	//float alpha;
    	int i;
    	String tmp = "";
    	//for (alpha = -6.0f; alpha <= 6.0f; alpha++)
    	//{
    	//	SynWTParamFilterFloatLift9x7 filter = new SynWTParamFilterFloatLift9x7(alpha);
    	//}
    	//alpha = -3.2f;
    	AnWTParamFilterFloatLift9x7 filter = new AnWTParamFilterFloatLift9x7(alpha);
    	float[] hpfilt = filter.getHPSynthesisFilter();
		float[] lpfilt = filter.getLPSynthesisFilter();
		float[] hpfilta = filter.getHPAnalysisFilter();
		float[] lpfilta = filter.getLPAnalysisFilter();
		
		float sum = 0f;
		tmp += "h = [";
		for (i = 0; i <lpfilta.length; i++)
		{
			tmp += lpfilta[i];
			sum += lpfilta[i];
			tmp += ", ";
		}
		tmp = tmp.substring(0, tmp.length() - 2);
		tmp += "]\nSUM=" + sum + "\n";

		sum = 0;
		tmp += "g = [";
		for (i = 0; i <hpfilta.length; i++)
		{
			tmp += hpfilta[i];
			sum += hpfilta[i];
			tmp += ", ";
		}
		tmp = tmp.substring(0, tmp.length() - 2);
		tmp += "]\nSUM=" + sum+ "\n";
		
		sum = 0;
		tmp += "rh = [";
		for (i = 0; i <lpfilt.length; i++)
		{
			tmp += lpfilt[i];
			sum += lpfilt[i];
			tmp += ", ";
		}
		tmp = tmp.substring(0, tmp.length() - 2);
		tmp += "]\nSUM=" + sum+ "\n";
		
		tmp += "rg = [";
		for (i = 0; i <hpfilt.length; i++)
		{
			tmp += hpfilt[i];
			sum += hpfilt[i];
			tmp += ", ";
		}
		tmp = tmp.substring(0, tmp.length() - 2);
		tmp += "]\nSUM=" + sum+ "\n";
		/*tmp += "figure\n title('Analysis, alpha = " + alpha + "')\n";
		tmp += "wavelet(h,g,5,0)\n";
		tmp += "figure\n title('Synthesis, alpha = " + alpha + "')\n";
		tmp += "wavelet(rh,rg,5,0)\n";*/		
		System.out.println (tmp);
	
 
    }
    
    public static void plotFilter(float alpha)
    {
    	AnWTParamFilterFloatLift9x7 filter = new AnWTParamFilterFloatLift9x7(alpha);
    	String fname =  "alpha" + alpha + ".gp";
    	filter.plotFilter("out/" + fname);
    }
    
    public static void plotFilter()
    {    	
    	try
		{			
			FileOutputStream fos = new FileOutputStream ("out/make_script");
			DataOutputStream out = new DataOutputStream(fos); 
			out.writeBytes("#!/bin/bash\n");
			int i = 100;
	    	for (float alpha = -6.0f; alpha <= -1.0f; alpha += 0.1f)
	    	{
	    		AnWTParamFilterFloatLift9x7 filter = new AnWTParamFilterFloatLift9x7(alpha);
	    		String fname =  i++ + "_alpha" + alpha + ".gp";     		
	    		filter.plotFilter("out/" + fname);	    	
	    		out.writeBytes ("echo \"set term png medium\n set xtics 1\n set ytics 1\n " +
	    				"set grid \n show grid\n  plot [] [-1:2] '" + fname + 
						"' title 'alpha = " +   Float.toString(((int)(alpha * 10))/10f) +  "' with linespoints \" " +
						"| gnuplot > " + fname.substring(0, fname.length()-3) + ".png\n");
	    	}	    	
	    	out.writeBytes ("convert -delay 20 *.png alpha.mng");
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}			    
    }
    
    public static void testBitPattern()
    {
    	SubbandStructure sb, sb2;
    	    	
    	
    	byte[] buf;
    	
    	sb = new SubbandStructure(512, 512, 0, 0, 3, null, null);
    	buf = sb.toBitPattern();    	
    	
    	sb2 = new SubbandStructure(512, 512, 0, 0, 3, null, null, buf);
    	sb2.plotTreeStructure("buf_tree.eps", 256, 256);
 
    	FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ArrayUtil.printBitArray(buf));
    }
    
    public static void testWSQ()
    {
    	byte[] wsq = ArrayUtil.createBitArrayFromHexString("f101248124803480148010");
    	FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ArrayUtil.printBitArray(wsq));
    }
    
    public static void testParamKey()
    {
    	String key;
    	//key = "65019188172819171891";    	
    	//key =   "82865662402565411568";
    	
    	//key =   "41403943424139394039"; //great key ;-)
    	//key = "00330066003366330066";
    	key = "66336666003333663366";
    	FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ParamKeyUtil.createFilterStringFromKey("w9x7p", key));
    	
    }
    
    public static void printAlphaSequenceFromKey(String key)
    {
    	FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ParamKeyUtil.createFilterStringFromKey("w9x7p", key));
    }
    
    public static void testPSNR()
    {
    	FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "PSNR: " + ImageStatsCalculator.calculatePSNR("lena512.pgm", "out.pgm"));
    }
    
    public static String createRandomKey(int l)
    {
    	Random r = new Random();
    	String key = "";
    	for (int i = 0; i < l; i++)
    	{
    		key += Integer.toString(r.nextInt(10));
    	}
    	System.out.println("Your key: " + key);
    	return key;
    }
    
    public static void testSobel()
    {
    	
    	Sobel sobel = new Sobel();
    	for (int id = 1; id <= 8; id++)
    	{
    		System.out.println ("id " + id + " angle " + sobel.IdtoReprAngl(id));
    	}
    	/*for (double angle =  -90;  angle <= 90; angle += 0.5)
    	{	
    		int id =  1+ (int) (8*  (angle + 89.9) / 180);
    		System.out.println ( "angle " + angle + " --> id : " + id);
    	}
		*/
		
		if (true) return; 
		
    	
    	try
		{
    		ImgReaderPGM reader = new ImgReaderPGM("lena512.pgm");
    		DataBlkInt blk =  ImageStatsCalculator.getDataBlkInt(reader,0);
    		//System.out.println (sobel.imageAngle(blk, 80));
    		//ImageStatsCalculator.writePGM(sobel.edgeImage(blk, 112), "test.pgm", 0);
    		System.out.println (sobel.imageReprAngleID(blk, 0));
    		
		}
    	catch (IOException e)
		{
    		System.out.println("ERROR");
		}
    	
    	
    	
    }
	
	public static void main(String[] args) {
		
		testWSQ();
		
		//printFilterCoefficients(3.7f);
		//String struct = "f100488400";
		//SubbandStructure sbbs = new SubbandStructure(512, 512, 0, 0, 5, null, null, ArrayUtil.createBitArrayFromHexString(struct));
		//String key = createRandomKey(20);
		//printAlphaSequenceFromKey("60378572082618643783");
		//String key = "14961807031883761415";
		//printAlphaSequenceFromKey(key);
		//testPSNR();
		//plotFilter();
		//testBitPattern();
		//testParamKey();
		//String hexString = "f124034806801834806801c01207801c01281240c34806801c01207801c01281240e01281240348087801c01281240e012812403480c12403480680100";		
		//byte[] buf = ArrayUtil.createBitArrayFromHexString(hexString);
		//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ArrayUtil.printBitArrayAsHex(buf));
		
		//printFilterCoefficients(-1.5861343420f);
		//printFilterCoefficients(6f);
		
		//printAlphaSequenceFromKey("");
		//plotFilter(-6.0f);
		//plotFilter(-2.3f);
		//plotFilter(2.3f);
		//plotFilter(-1.5861343420f);
		
		/*System.out.println (Long.MIN_VALUE + " -- " + Long.MIN_VALUE);
		Random random = new Random();
		for (int i = 0; i < 10; i ++)
		{
			System.out.println (i + ": " + random.nextInt(256));
		}*/
		
		
	}
}
