package jj2000.j2k.codestream.reader;

import java.util.Iterator;
import java.util.Vector;

import jj2000.j2k.codestream.reader.CBlkInfo.InclusionInfo;

public class DataCollector {

	/** Information on all packets and the associated cblks */
	public Vector<PktCblksInfo> pktCblkInfos;	
	
	/**
	 * create a data collector. this can be used to encrypt header data and write a new 
	 * file or to decrypt a previously encrypted file
	 * 
	 * @param ehs the input stream 
	 * @param maxRL the maximum number of resolution levels
	 * @param seed seed for the random number generator
	 * @param pl the parameter list
	 */
	public DataCollector() {
		pktCblkInfos = new Vector<PktCblksInfo>();
		//allCblks = new Vector<CBlkInfo>();
	}
	
	/**
	 * add information regarding a specific packet
	 * 
	 * @param pci the packet-codeblock info
	 */
	public void addPktCblkInfo(PktCblksInfo pci) {
		pktCblkInfos.add(pci.getPacketIdx(), pci);
	}
	
	/**
	 * get the packet-codeblock info object for packet with
	 * index pktIdx
	 * 
	 * @param pktIdx packet index
	 * @return packet-codeblock info
	 */
	public PktCblksInfo getPktCblkInfo(int pktIdx) {
		return pktCblkInfos.get(pktIdx);
	}

	/**
	 * get all collected info on all packets
	 * 
	 * @return packet-cblk info for all packets
	 */
	public Vector<PktCblksInfo> getPktCblkInfos() {
		return pktCblkInfos;
	}

	/**
	 * return a vector with all packets in resolution r
	 * 
	 * @param r  resolution level           
	 * @return packets in r
	 */
	public Vector<PktCblksInfo> getPacketsInResolution(int r) {
		Vector<PktCblksInfo> packetsInRes = new Vector<PktCblksInfo>();
		Iterator<PktCblksInfo> it = pktCblkInfos.iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			if (pci.getResolution() == r) {
				packetsInRes.add(pci);
			}
		}
		return packetsInRes;
	}
	
	/**
	 * return a vector with all packets in resolution r
	 * 
	 * @param r  resolution level           
	 * @return packets in r
	 */
	public Vector<PktCblksInfo> getPacketsInResolution(int r, int sIdx) {
		Vector<PktCblksInfo> packetsInRes = new Vector<PktCblksInfo>();
		Iterator<PktCblksInfo> it = pktCblkInfos.iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			if (pci.getResolution() == r && pci.getPacketIdx() >= sIdx) {
				packetsInRes.add(pci);
			}
		}
		return packetsInRes;
	}


	/**
	 * get all codeblocks in resolution r
	 * @param r resolution level
	 * @return codeblocks in r
	 */
	public Vector<CBlkInfo> getCodeBlocksInRes (int r) {
		Vector<CBlkInfo> cblks = new Vector<CBlkInfo>();
		PktCblksInfo pci  = getPacketsInResolution(r).get(0);			
		Iterator <CBlkInfo> it = pci.getAllCBlks().iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			cblks.add(ccb);
		}
		return cblks;
	}
	
	/**
	 * get all codeblocks in resolution r
	 * @param r resolution level
	 * @return codeblocks in r
	 */
	public Vector<CBlkInfo> getCodeBlocksInRes (int r, int sIdx) {
		Vector<CBlkInfo> cblks = new Vector<CBlkInfo>();
		Iterator<PktCblksInfo> pit  = getPacketsInResolution(r, sIdx).iterator();
		while (pit.hasNext()) {
			PktCblksInfo pci = pit.next();
			cblks.addAll(pci.getCBlksByInclusion(InclusionInfo.FI));
		}
		return cblks;
	}

	

	
	/** get all codeblocks of all resolutions */
	public Vector<CBlkInfo> getAllCodeBlocks () {
		Vector <CBlkInfo> cblks = new Vector<CBlkInfo>();
		int r = 0;
		
		while (getPacketsInResolution(r).size() > 0) {
			cblks.addAll(getPacketsInResolution(r).
					get(0).getAllCBlks());
			r++;
		}		
		return cblks;		
	}
	
}
