/*
 * Created on Mar 13, 2007
 *
 */
package jj2000.j2k.codestream.reader;

import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import jj2000.j2k.codestream.reader.CBlkInfo.InclusionInfo;
import jj2000.j2k.codestream.writer.FileCodestreamWriter;
import jj2000.j2k.io.RandomAccessIO;

/**
 * This class collects information on JPEG2000 code packets and the
 * codeblocks associated with these packets. 
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 *
 */
public class PktCblksInfo implements Comparable<PktCblksInfo>{
	
    /** Index of the packet */
    private int packetIdx;
    
    /** Resolution of packet */
    private int resolution;
    
    /** Layer of pakcet */
    private int layer;
        
    private int packetOffset;
    
    private int packetDataOffset;
    
    private int packetBodyLength;
    
    private byte[] header;
    
    private byte[] body;
    
    private byte[] fakedHeader;       
    
    private Vector<CBlkInfo> all_cblks;
    
    private Vector<InclusionInfo> savedInclusionVector;
     
                
    /**
     * create new info object for codeblocks in a packet 
     * @param resolution
     * @param layer
     */
    public PktCblksInfo (int packetIdx, int resolution, int layer) {
    	this.packetIdx = packetIdx;
    	this.resolution = resolution;
    	this.layer = layer;
    	all_cblks = new Vector<CBlkInfo>();
    }

    /**
     * the resolution of this packet
     * @return
     */
	public int getResolution() {
		return resolution;
	}

	/** the layer of this packet 
	 * 
	 * @return
	 */
	public int getLayer() {
		return layer;
	}

	/** packetindex for this packet */
	public int getPacketIdx() {
		return packetIdx;
	}
		
	public Vector<InclusionInfo> getInclusionVector () {
		Vector<InclusionInfo> inc = new Vector<InclusionInfo>();
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			inc.add(ccb.inclusionInfo[layer]);
		}
		return inc;
	}
	
	public Vector<InclusionInfo> getInclusionVectorFINI () {
		Vector<InclusionInfo> inc = new Vector<InclusionInfo>();
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (ccb.inclusionInfo[layer] == InclusionInfo.NI ||
					ccb.inclusionInfo[layer]  == InclusionInfo.FI) {
				inc.add(ccb.inclusionInfo[layer]);
			}
		}
		return inc;
	}
	
	public Vector<InclusionInfo> getInclusionVectorPIPN () {
		Vector<InclusionInfo> inc = new Vector<InclusionInfo>();
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (ccb.inclusionInfo[layer] == InclusionInfo.PI ||
					ccb.inclusionInfo[layer]  == InclusionInfo.PN) {
				inc.add(ccb.inclusionInfo[layer]);
			}
		}
		return inc;
	}
	
	
	/**
	 * read the body-data for this packet from <tt>ehs</tt> 
	 * into the private buffer
	 *  
	 * @param ehs the RandomAccessIO to read from
	 */
	public void readBody (RandomAccessIO ehs) {
		try {
			int saveOff = ehs.getPos();
			body = new byte[packetBodyLength];
			ehs.seek(packetDataOffset);
			ehs.readFully(body, 0, packetBodyLength);
			ehs.seek(saveOff);
		} catch (IOException ioe) {
			System.err.println ("Could not read body data for " + toString() + "\n Error:" + ioe.toString());
			System.exit(1);
		}
		
	}
	
	
	/** read the original head of this packet
	 * 
	 * @param ehs Randomaccessio to read from
	 */
	public void readHead(RandomAccessIO ehs) {		
		try {
			int saveOff = ehs.getPos();
			int hlen = packetDataOffset -  packetOffset;
			//reading of header not correct for EPH used (because it also reads EPH!!
			header = new byte[hlen];
			ehs.seek(packetOffset);
			ehs.readFully(header, 0, hlen);
			ehs.seek(saveOff);
		} catch (IOException ioe) {
			System.err.println ("Could not read header data for " + toString() + "\n Error:" + ioe.toString());
			System.exit(1);
		}

		
	}
	
	/** 
	 * the saved body data for this packet 
	 * @return body data 
	 */
	public byte[] getBody() {
		return body;
	}
	
	/**
	 * write the original packet to writer
	 * @param writer
	 * @throws IOException
	 */
	public void writePacket (FileCodestreamWriter writer, boolean sop, boolean eph) throws IOException {
		writer.writePacketHead(header, getHeaderLength(), false, sop, eph);
		writer.writePacketBody(body, packetBodyLength, false, false, 0);
	}


	/**
	 * writer the faked packet to writer 
	 * 
	 * @param writer
	 * @throws IOException
	 */
	public void writeFakedPacket (FileCodestreamWriter writer, boolean sop, boolean eph) throws IOException {
		writer.writePacketHead(fakedHeader, fakedHeader.length, false, sop, eph);
		writer.writePacketBody(body, packetBodyLength, false, false, 0);
	}
	
	/**
	 * simluate body encryption
	 *
	 */
	public void simulateBodyEncryption () {
		Random r = new Random();
		
		for (int i = 0; i < body.length; i++) {
			body[i] = (byte) (r.nextInt(0x8f)); 
		}	
	}

	
	/*public int getLenFirst() {
		if (lenFirst == -1) { 
			lenFirst = 0;
			CBlkInfo ccb;
			Iterator<CBlkInfo> it = cblks_first_inc.iterator();
			while (it.hasNext()) {
				ccb =  it.next();
				//CBlkInfo ccb = cbI[0][pci.getResolution()][cie.getSubband()][cie.getCblkCoord().x][cie.getCblkCoord().y];
				lenFirst += ccb.len[layer]; 
			}			
		}
		return lenFirst;
	}*/
	
	/*public int getLenPrev() {
		if (lenPrev == -1) { 
			lenPrev = 0;
			CBlkInfo ccb;
			Iterator<CBlkInfo> it = cblks_prev_inc_and_inc.iterator();
			while (it.hasNext()) {
				ccb = it.next();
				//CBlkInfo ccb = cbI[0][pci.getResolution()][cie.getSubband()][cie.getCblkCoord().x][cie.getCblkCoord().y];
				lenPrev += ccb.len[layer]; 
			}			
		}
		return lenPrev;
		
	}*/
	
	/*public Vector<CBlkInfo> getAllCblks () {
		Vector<CBlkInfo> all = new Vector<CBlkInfo>();
		all.addAll(cblks_first_inc);
		all.addAll(cblks_prev_inc_and_inc);
		all.addAll(cblks_not_inc);
		all.addAll(cblks_prev_inc_and_not_inc);
		Collections.sort(all);
		return all;
	}*/
	
	public Vector<CBlkInfo> getCBlksByInclusion (InclusionInfo ii) {
		Vector<CBlkInfo> res = new Vector<CBlkInfo>();
		
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (ccb.inclusionInfo[layer] == ii) {
				res.add(ccb);
			}
		}
		//Collections.sort(res); //actually, it should be sorted already!
		return res;		
	}
	
	
	public Vector<CBlkInfo> getCblksInSubband (int sbIdx) {
		Vector<CBlkInfo> res = new Vector<CBlkInfo>();
		
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (ccb.sbIdx== sbIdx) {
				res.add(ccb);
			}
		}
		//Collections.sort(res); //actually, it should be sorted already!
		return res;		
	}
	
	public int getDataLenByInclusion (InclusionInfo ii) {
		int sum = 0;
		
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (ccb.inclusionInfo[layer] == ii) {
				sum += ccb.len[layer];
			}
		}
		//Collections.sort(res); //actually, it should be sorted already!
		return sum;		
	}
	
	
		
	public Vector<CBlkInfo> getNonEmptyCblks () {
		Vector<CBlkInfo> nec = new Vector<CBlkInfo>();
		nec.addAll(getCBlksByInclusion(InclusionInfo.FI));
		nec.addAll(getCBlksByInclusion(InclusionInfo.PI));		
		Collections.sort(nec);		
//		System.out.println ("SORTED: " + nec.toString());
		return nec;
	}
	
	public Vector<CBlkInfo> getEmptyCblks () {
		Vector<CBlkInfo> ec = new Vector<CBlkInfo>();
		ec.addAll(getCBlksByInclusion(InclusionInfo.NI));
		ec.addAll(getCBlksByInclusion(InclusionInfo.PN));		
		Collections.sort(ec);		
//		System.out.println ("SORTED: " + nec.toString());
		return ec;
	}
	
	/**
	 * set the inclusion info for the codeblocks in this packet according
	 * to <tt>incVector</tt>. Offsets, lenghts, and number of truncation points
	 * will be updated. 
	 *  
	 * @param incVector new incluson information
	 */
	public void setInclusionVector (Vector<InclusionInfo> incVector, boolean decode) {
		/*
		 * 		Vector<InclusionInfo> inc = new Vector<InclusionInfo>();
		Iterator<CBlkInfo> it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			inc.add(ccb.inclusionInfo[layer]);
		}
		return inc;			
		 */
		
		// first we get the original incl. info
		Vector<CBlkInfo> realEmptyBlocks = getEmptyCblks();
		Vector<CBlkInfo> realNonEmptyBlocks = getNonEmptyCblks();
		Vector<CBlkInfo> realFIs = getCBlksByInclusion(CBlkInfo.InclusionInfo.FI);

		//now we set the new inclusion info!
		Iterator<InclusionInfo> itIncInfo = incVector.iterator();
		Iterator<CBlkInfo> itCblks = all_cblks.iterator();
		Iterator<CBlkInfo> itEmptyCblks = realEmptyBlocks.iterator();
		Iterator<CBlkInfo> itNonEmptyCblks = realNonEmptyBlocks.iterator();
		Iterator<CBlkInfo> itFIs = realFIs.iterator();
		
		while (itIncInfo.hasNext()) {
			InclusionInfo ii = itIncInfo.next();
			if (!itCblks.hasNext()) {
				System.err.println ("Inclusion Vector is too long, sorry.");
				System.exit(1);
			}
			CBlkInfo ccb = itCblks.next();
			ccb.inclusionInfo[layer]  = ii;
			
			// the real cblk info as reference
			CBlkInfo ccb_ref = null;
			if (ii == InclusionInfo.FI || 
					ii == InclusionInfo.PI) {
				if ( !itNonEmptyCblks.hasNext() ) {
					System.err.println ("Not enough non-empty reference blocks in real inclusion info found, sorry.");					
					System.exit(1);					
				}
				ccb_ref = itNonEmptyCblks.next();
			} else {
				if ( !itEmptyCblks.hasNext() ) {
					System.err.println ("Not enough non-empty reference blocks in real inclusion info found, sorry.");					
					System.exit(1);					
				}
				ccb_ref = itEmptyCblks.next();			
			}
						
			ccb.len[layer] = ccb_ref.savedCblk.len[layer];
			ccb.off[layer] = ccb_ref.savedCblk.off[layer];
			ccb.ntp[layer] = ccb_ref.savedCblk.ntp[layer];
			
			if (ii == CBlkInfo.InclusionInfo.FI) {
				if (! itFIs.hasNext()) {
					System.err.println ("Not enough FI reference blocks in real inclusion info found, sorry.");					
					System.exit(1);										
				}
				CBlkInfo cb_fi_ref = itFIs.next();
				ccb.msbSkipped = cb_fi_ref.savedCblk.msbSkipped;
			}
			
			/*if (decode && ii == CBlkInfo.InclusionInfo.FI)
				ccb.msbSkipped = ccb_ref.savedCblk.msbSkipped;
			if (!decode && ccb_ref.savedCblk.inclusionInfo[layer] == CBlkInfo.InclusionInfo.FI) {
				ccb.msbSkipped = ccb_ref.savedCblk.msbSkipped;
			}*/
			//ccb.sbIdx = ccb_ref.savedCblk.sbIdx;
			ccb.updateCTP();
			
			//System.out.print("Block "  + ccb.coord + " " + ccb.savedCblk.inclusionInfo[layer] + "-->" +ccb.inclusionInfo[layer]);
			//System.out.print ("; len: " + ccb.savedCblk.len[layer] + "-->" + ccb.len[layer] + "; off: " + ccb.savedCblk.off[layer] + "-->" + ccb.off[layer] + "; ntp: " + ccb.savedCblk.ntp[layer] + "-->" + ccb.ntp[layer]);
			//System.out.println ( "; ctp: " + ccb.savedCblk.ctp + " --> " + ccb.ctp);
			
			/*if (ccb.pktIdx[layer] != ccb_ref.realCblk.pktIdx[layer]) {
				System.err.println (" Something is different than Dominik thought.");
				System.exit(1);
			}
			//this should not be necessary
			ccb.pktIdx[layer] = ccb_ref.realCblk.pktIdx[layer];*/		
		}		
	}
	
	/**
	 * for each recorded codeblock, save it's state 
	 * in the CBlkInfo-Object
	 *
	 */
	public void saveCblkState () {
		Iterator<CBlkInfo> it = getAllCBlks().iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			ccb.savedCblk = new CBlkInfo(ccb);
		}		
	}

	/**
	 * offset of this packet's body  in the stream
	 * @return
	 */
	public int getPacketDataOffset() {
		return packetDataOffset;
	}
	
	public void setPacketDataOffset(int packetDataOffset) {
		this.packetDataOffset = packetDataOffset;
	}


	/** offset of this packet in the stream */
	public int getPacketOffset() {
		return packetOffset;
	}

	public void setPacketOffset(int packetOffset) {
		this.packetOffset = packetOffset;
	}

	public int getHeaderLength () {
		return packetDataOffset - packetOffset;
	}

	public int getPacketBodyLength() {
		return packetBodyLength;
	}

	public void setPacketBodyLength(int packetBodyLength) {
		this.packetBodyLength = packetBodyLength;
	}
	
	public String toString() {
		String string = "";
		string += "# " + packetIdx + " l=" +layer + " r="+resolution + " data size " + packetBodyLength + " empty cblk: " + getEmptyCblks().size() + " non-empty cblks: " + getNonEmptyCblks().size();
		return string;
	}

	/**
	 * get all codeblocks in this packet (non-empty and empty)
	 * 
	 * @return all codeblocks in this packet
	 */
	public Vector<CBlkInfo> getAllCBlks() {
		return all_cblks;
	}
	
	public void addCblk (CBlkInfo cblk) {
		all_cblks.add(cblk);
	}

	/** get the saved inclusion vector stored for this object
	 * 
	 * @return saved inclusion vector
	 */
	public Vector<InclusionInfo> getSavedInclusionVector() {
		return savedInclusionVector;
	}

	/** save an inclusion vector in this object
	 *  
	 * @param savedInclusionVector
	 */
	public void setSavedInclusionVector(Vector<InclusionInfo> savedInclusionVector) {
		this.savedInclusionVector = savedInclusionVector;
	}
	

	public byte[] getOriginalHeader() {
		return header;
	}


	public byte[] getFakedHeader() {
		return fakedHeader;
	}

	public void setFakedHeader(byte[] fakedHeader) {
		this.fakedHeader = fakedHeader;
	}
	
	/**
	 * sort by res, layer
	 */
	public int compareTo (PktCblksInfo other ) {
		if (this.resolution < other.resolution) {
			return -1;
		} else if (this.resolution > other.resolution) {
			return 1;
		} else if (this.layer < other.layer) {
			return -1;
		} else if (this.layer > other.layer) {
			return 1; 
		} else //same packet
			return 0;
	}
	

}
