package jj2000.j2k.encryption;

import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

public class HeaderHashCompare {
	
	/**
	 * compare all hashes in the first vector to all hashes in the
	 * second vector. the individual results are written to fname
	 * 
	 * @param hashes1
	 * @param hashes2
	 * @param fname
	 */
	public void writeComparisonToFile (Vector<Vector<Integer>> hashes1,  Vector<Vector<Integer>> hashes2, String fname) {
		
		if (hashes1.size() != hashes2.size() ) {
			System.err.println ("Need same number of hashes to compare on both sides.");
			System.exit (1);
		}
		
		try
		{
						
			FileOutputStream fos = new FileOutputStream (fname, true);
			DataOutputStream out = new DataOutputStream(fos);

			for (int i = 0; i < hashes1.size(); i++) {
				out.writeBytes ( " " + compareHash(hashes1.get(i), hashes2.get(i)));				
			}
			out.writeBytes("\n");

		
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
		finally
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Header Hash Comparison written to " + fname);	
		
		}	
	}
	
	/**
	 * compare two hashes and return the number of positions where they differ
	 * @param hash1
	 * @param hash2
	 * @return number of differing positions
	 */
	public float compareHash (Vector<Integer> hash1, Vector<Integer>  hash2) {
		int samecount = 0;
		int nonzerocount = 0;
		
		if (hash1.size() != hash2.size() ) {
			System.err.println ("Sizes of hashes differ (" + hash1.size() + ":" + hash2.size() + ")" );
			return -1;
		}
		

		
		for (int i = 0; i < hash1.size(); i++) {			
			if  (hash1.get(i) != 0 || hash2.get(i) != 0) {				
				nonzerocount++;				
				if (hash1.get(i).intValue() == hash2.get(i).intValue()) {
					//System.out.println (i + " Hash 1 " + hash1.get(i) + " -- Hash 2 " + hash2.get(i));
					samecount++;
				}
			}
		}
		
		return ((float)samecount)/((float)nonzerocount);
	}

}
