package jj2000.j2k.encryption;

import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.Iterator;
import java.util.Vector;

import jj2000.j2k.codestream.reader.CBlkInfo;
import jj2000.j2k.codestream.reader.DataCollector;
import jj2000.j2k.codestream.reader.PktCblksInfo;
import jj2000.j2k.codestream.reader.CBlkInfo.InclusionInfo;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

public class HeaderHasher {
	
	private Vector<Integer> iInfo;
	private Vector<Integer> msbs;
	private Vector<Integer> lengths;
	private Vector<Integer> ncps;
	
	private DataCollector dc;
	
	public HeaderHasher (DataCollector dc) {
		this.dc = dc;

	}
	
	
	/** 
	 * Write all hashes to file <tt>fname</tt>
	 */
	public void hashToFile(String fname) {		
		HeaderHashFileWriter writer = new HeaderHashFileWriter();
		writer.writeHashes(getAllHashes(), fname);
	}
	

	public Vector<Vector<Integer>> getAllHashes () {
		Vector<Vector<Integer>> vec = new Vector<Vector<Integer>>();
		vec.add (getInclusionVector());
		vec.add (getMSBVector());
		vec.add (getLengthsVector());
		vec.add (getNCPVector());
		return vec;
	}
	
	
	
	
	private void  createInclusionVector () {

		iInfo = new Vector<Integer>(); 
		//sort by res, layer
		Collections.sort (dc.getPktCblkInfos());
		
		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			Iterator<InclusionInfo> iit = pci.getInclusionVector().iterator();
			while (iit.hasNext()) {
				iInfo.add(iit.next().ordinal() + 1); //don't use 0, b/c it
				// is treated differently when hashes are compared
			}						
		}
				
	}
	
	
	private void createMSBVector () {
		Vector<CBlkInfo> cblks = dc.getAllCodeBlocks();
		msbs = new Vector<Integer>();
		
		Collections.sort(cblks);
		
		Iterator<CBlkInfo> it = cblks.iterator();
		while (it.hasNext()) {
			msbs.add( (it.next()).msbSkipped + 1);  //don't use 0, b/c it
			// is treated differently when hashes are compared
		}
		
	}
		
	private void createLengthsVector () {

		lengths = new Vector<Integer>(); 
		//sort by res, layer
		Collections.sort (dc.getPktCblkInfos());
		
		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			Iterator<CBlkInfo> cit = pci.getAllCBlks().iterator();
			while (cit.hasNext()) {
				lengths.add (cit.next().len[pci.getLayer()]);
			}
		}
	
	}

	
	private void createNCPVector () {

		ncps = new Vector<Integer>(); 
		//sort by res, layer
		Collections.sort (dc.getPktCblkInfos());
		
		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			Iterator<CBlkInfo> cit = pci.getAllCBlks().iterator();
			while (cit.hasNext()) {
				ncps.add (cit.next().ntp[pci.getLayer()]);
			}
		}
	
	}
	
	public Vector<Integer>getInclusionVector() {
		if (iInfo == null) createInclusionVector();
		return iInfo;
	}
	
	public Vector<Integer>getMSBVector() {
		if (msbs == null) createMSBVector();
		return msbs;
	}

	public Vector<Integer>getLengthsVector() {
		if (lengths == null) createLengthsVector();
		return lengths;
	}

	public Vector<Integer>getNCPVector() {
		if (ncps == null) createNCPVector();
		return ncps;
	}	
	
	

}
