/* 
 * CVS identifier:
 * 
 * $Id: ImageStatsCalculator.java,v 1.21 2006-06-29 09:58:53 dengel Exp $
 * 
 * Class:                   ImageStatsCalculator
 * 
 * Description:             
 * 
 * Created on Nov 18, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */

package jj2000.j2k.image;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;



import jj2000.disp.BlkImgDataSrcImageProducer;
import jj2000.disp.ImgPane;
import jj2000.j2k.image.edgedectection.Sobel;
import jj2000.j2k.image.input.ImgReaderPGM;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;


/**
 * @author dengel
 * 
 * this class provides a collection of static methods that can be applied to 
 * sources of type BlkImgDataSrc
 *
 */
public class ImageStatsCalculator {
	
	
	
	private static double log (double num, double base)
	{
		return   Math.log (num) /  Math.log (base);
	}

	
	private static double log10 (double num)
	{
		return log (num, 10);
	}



	
	/**
	 *  * Calculates the PSNR wit peak_to_peak= 256<br><br>
	 * 
	 *  NOTE that this function does not regard components or tilings!
	 * 
	 * @param org_src 
	 * @param new_src
	 * @return PSNR
	 */
	public static double calculatePSNR (BlkImgDataSrc org_src , BlkImgDataSrc new_src)
	{		
		return calculatePSNR(org_src, new_src, 256);
		
	}
	
	
	
	/**
	 * Calculate the PSNR between two filse
	 * 
	 * @param org_file_name
	 * @param new_file_name
	 * @return psnr
	 */
	public static double calculatePSNR (String org_file_name, String new_file_name)
	{
		double psnr = 0;
		try
		{
			ImgReaderPGM org_reader = new ImgReaderPGM(org_file_name);
			ImgReaderPGM new_reader = new ImgReaderPGM(new_file_name);
			psnr = calculatePSNR((BlkImgDataSrc) org_reader, (BlkImgDataSrc) new_reader); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}
		
		
		return psnr;
	}
		
	
	
	/**
	 * Calculate the PSNR between a file containgin the original image and an BlkImgDataSrc that represents the reconstructed image
	 * 
	 * @param file_name
	 * @param new_src
	 * @return psnr
	 */
	public static double calculatePSNR (String file_name, BlkImgDataSrc new_src)
	{
		double psnr = 0;
		try
		{
			ImgReaderPGM reader = new ImgReaderPGM(file_name);
			psnr = calculatePSNR((BlkImgDataSrc) reader, new_src); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}
		
		
		return psnr;
	}
		
	
	/**
	 * 
	 * Calculates the PSNR with specifying peak_to_peak<br><br>
	 * 
	 * NOTE that this function does not regard components or tilings!
	 * 
	 * @param org_src the original image
	 * @param new_src the image with which org_src is to be compared
	 * @param peak_to_peak Max peak to peak
	 * @return PSNR
	 */
	public static double calculatePSNR (BlkImgDataSrc org_src , BlkImgDataSrc new_src , int peak_to_peak)
	{
		
	/*
	 * double Image::CalcPSNR(const Image& img, int peak_to_peak=256) const
	 	{
	 		return 20.0 * log10 (peak_to_peak / sqrt (MeanSquaredError (img)));
	 	}
	 */
	return 20.0d * log10( (double) peak_to_peak / Math.sqrt(calculateMSE(org_src, new_src)));			
	}
	
	

	/**
	 * 
	 * Calculates the MSE from two image sources
	 * 
	 * NOTE that this function does not regard components or tilings!
	 * 
	 * @param org_src the original image
	 * @param new_src the image with which org_src is to be compared
	 * @param c the component no. in both images to be used
	 * @return MSE
	 */
	
	public static double  calculateMSE (BlkImgDataSrc org_src , BlkImgDataSrc new_src)
	{
		
		DataBlkInt org_data_blk = getDataBlkInt(org_src, 0);
		DataBlkInt new_data_blk = getDataBlkInt(new_src, 0);
		
		if ( org_data_blk.w != new_data_blk.w ||  
			 org_data_blk.h != new_data_blk.h)
		{
			throw (new IllegalArgumentException("Images are not of the same dimension"));
		}
		
		
		int ulx = org_data_blk.ulx;
		int uly = org_data_blk.uly;
		int w = org_data_blk.w;
		int h = org_data_blk.h;
		
		int org_data  [] = org_data_blk.data;
		int new_data  [] = new_data_blk.data;
		
		
		int i, j ;
		double overallMSE = 0.0d;
		
		//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "org nombits " + org_src.getNomRangeBits(0));
		//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "new nombits " + new_src.getNomRangeBits(0));
		
		
		int nd = 0;
		int od = 0;
		for (j = 0; j < h; j++)
		{
			int rowMSE = 0;
			for(i = 0; i < w; i++)
			{

				od = org_data[j*w + i] + 128; //add DC_OFFSET
				nd = new_data[j*w + i] + 128; //add DC_OFFSET
				nd = (nd < 0) ? 0 : (nd > 255) ? 255 : nd; //check range
				
				int error = od - nd;

				rowMSE +=  error * error;				
			}
			overallMSE += (double) rowMSE / (double) w;			
		}		
		return overallMSE / (double) h;
		
		 /* 
		  * taken from Ganesh ++
		  *  
		
		double Image::MeanSquaredError (const Image &I) const
		{
		  if ((GetWidth() != I.GetWidth()) || (GetHeight() != I.GetHeight()))
		     throw RuntimeError ("Image sizes are different!");
		
		  double OverallMSE = 0.0;
		  unsigned Width = GetWidth ();
		  unsigned Height = GetHeight ();
		  for (unsigned y = 0; y < Height; y ++)
		  { unsigned RowMSE = 0;
		    for (unsigned i = y * Width; i < (y + 1) * Width; i ++)
		    {   int Error = operator() (i) - I (i);
		        RowMSE += Error * Error;
		    }
		    OverallMSE += (double) RowMSE / (double) Width;
		  }
		  return OverallMSE / (double) GetHeight ();
		}
		
		*/

		
	}
	
	
	
	public static double calculateAverage(BlkImgDataSrc src)
	{
		DataBlkInt data_blk = getDataBlkInt(src, 0);
		int data  [] = data_blk.data;
		int d = 0;
		int w = data_blk.w;
		int h = data_blk.h;
		double average = 0;
				
		for (int j = 0; j < h; j++)
		{		
			int row_sum = 0;
			for(int i = 0; i < w; i++)
			{
				d = data[j*w + i] + 128; //add DC_OFFSET
				d = data[j*w + i] + 128; //add DC_OFFSET
				d = (d < 0) ? 0 : (d > 255) ? 255 : d; //check range	
				row_sum += d;
			}			
			average += (double)row_sum/ (double)w;					
		}		
		average = average / (double) h;
		
		//testing
		//writePGM(src, "test.pgm", 0);
		
		return average;		
	}
	
	/**
	 * Calculates the variance in a reconstructed image
	 * 
	 * @param src
	 * @return variance of src
	 */
	public static double calculateVariance(BlkImgDataSrc src)
	{
		DataBlkInt data_blk = getDataBlkInt(src, 0);
		int data  [] = data_blk.data;
		int d = 0;
		int w = data_blk.w;
		int h = data_blk.h;
		
		double avg = calculateAverage(src);
		double var = 0.0;
		
		//System.out.println ("Avg " + avg);
				
		for (int j = 0; j < h; j++)
		{
			double row_var = 0.0;
			for(int i = 0; i < w; i++)
			{
				d = data[j*w + i] + 128; //add DC_OFFSET
				d = data[j*w + i] + 128; //add DC_OFFSET
				d = (d < 0) ? 0 : (d > 255) ? 255 : d; //check range
				
				row_var += Math.pow((d - avg), 2);
			}						
			var += row_var / (double) w;
		}		
		var = var / (double) h;
		
		//testing
		//writePGM(src, "test.pgm", 0);
		
		return var;		
	}
	
	
	/**
	 * get the data from an image source as a block of integer data
	 * 
	 * @param src original image
	 * @param c component number
	 * @return datablock
	 */
	public static DataBlkInt getDataBlkInt (BlkImgDataSrc src, int c) 
	{
		//	ImgDataConverter src = new ImgDataConverter(_src); 
		// Check that imgSrc is of the correct type
        // Check that the component index is valid
        if( (c<0) || ( c>=src.getNumComps() ) || 
            ( src.getNomRangeBits(c)>8 ) )
            throw new IllegalArgumentException();
        
		DataBlkInt data_blk;
		
		int ulx = src.getImgULX();
		int uly = src.getImgULY();
		int w = src.getImgWidth();
		int h = src.getImgHeight();
	    //int fb = src.getFixedPoint(c);
	    //int levShift = 1<< (src.getNomRangeBits(c)-1);
	    //int tmp,maxVal= (1<<src.getNomRangeBits(c))-1;
	    
	    
		data_blk = new DataBlkInt(ulx, uly, w, h);
				
	    // Check the array size
        if (data_blk.data != null && data_blk.data.length < w*h) {
            // A new one will be allocated by getInternCompData()
            data_blk.data = null;
        }
		
        // Request the data and make sure it is not
        // progressive
        do {
            data_blk = (DataBlkInt)src.getInternCompData(data_blk,c);
        } while (data_blk.progressive);				
		
        return data_blk;
	
	}
	

	
	/**
	 * Returns the Luminance Similarity Score (LSS) as described in Mao04a, with reasonable default parameters
	 * 
	 * @param org_src Original image
	 * @param new_src Reconstructed image
	 * @return LSS
	 */
	public static double calculateLSS (String org_file_name, String new_file_name)
	{
		double alpha = 0.1;
		double beta = 3;
		int blocksize = 8;

		double LSS = 0;
		try
		{
			ImgReaderPGM org_reader = new ImgReaderPGM(org_file_name);
			ImgReaderPGM new_reader = new ImgReaderPGM(new_file_name);
			LSS = calculateLSS((BlkImgDataSrc) org_reader, (BlkImgDataSrc) new_reader, alpha, beta, blocksize); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}			
		return LSS; 
	}


	/**
	 * 
	 * Returns the Luminance Similarity Score (LSS) as described in Mao04a
	 * 
	 * @param file_name original image
	 * @param new_src reconstructed image
	 * @param alpha see Mao04a
	 * @param beta see Mao04a
	 * @return lss
	 */
	public static double calculateLSS (String file_name, BlkImgDataSrc new_src, double alpha, double beta, int blocksize)
	{
		double LSS = 0;
		try
		{
			ImgReaderPGM reader = new ImgReaderPGM(file_name);
			LSS = calculateLSS((BlkImgDataSrc) reader, new_src, alpha, beta, blocksize); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}			
		return LSS; 
	}		
	
	
	/**
	 * Returns the Luminance Similarity Score (LSS) as described in Mao04a
	 * 
	 * @param org_src Original image
	 * @param new_src Reconstructed image
	 * @param alpha factor alpha, see Mao04a
	 * @param beta quantization paramter beta, see Mao04a
	 * @return LSS
	 */
	public static double calculateLSS (BlkImgDataSrc org_src, BlkImgDataSrc new_src, double alpha, double beta, int blocksize)
	{
		DataBlkInt org_data_blk = getDataBlkInt(org_src, 0);
		DataBlkInt new_data_blk = getDataBlkInt(new_src, 0);
		
		
		if ( org_data_blk.w != new_data_blk.w ||  
			 org_data_blk.h != new_data_blk.h)
		{
			//throw (new IllegalArgumentException("Images are not of the same dimension"));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "LSS/ESS: Images are not of the same dimension. LSS not calculated.");
			return 0;
		}
	 				
		int ulx = org_data_blk.ulx;
		int uly = org_data_blk.uly;
		int w = org_data_blk.w;
		int h = org_data_blk.h;
		
		if (w % blocksize != 0 || h % blocksize != 0)
		{
			//throw (new IllegalArgumentException("Image dimension cannot be divided by blocksize. (Not implemented)"));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "LSS/ESS: Image dimension cannot be divided by blocksize. (Not implemented). LSS not calculated.");
			return 0;
		}
		
		int org_data  [] = org_data_blk.data;
		int new_data  [] = new_data_blk.data;
		
		
		int i, j ;
		double LSS = 0.0d;
				
		int nd = 0;
		int od = 0;
		for (j = 0; j < h; j+=blocksize)
			for(i = 0; i < w; i+=blocksize)
			{
				int blockLSS = 0;
				int avgLum = 0;
				double o_avg = 0d;
				double n_avg = 0d;
			    for(int jj=0; jj<blocksize; jj++)
			    	for (int ii=0; ii<blocksize; ii++)
			    	{
						od = org_data[(j+jj)*w + i+ii] + 128; //add DC_OFFSET
						nd = new_data[(j+jj)*w + i+ii] + 128; //add DC_OFFSET
						
						nd = (nd < 0) ? 0 : (nd > 255) ? 255 : nd; //check range						
						
						o_avg += od;
						n_avg += nd;						
			    	}
		    	o_avg = o_avg / (Math.pow(blocksize, 2));
		    	n_avg = n_avg / (Math.pow(blocksize, 2));
		    	double error = Math.abs (o_avg - n_avg);
				if (error < beta / 2d)
				{
					LSS +=  1;
				}
				else
				{
					LSS += -alpha * Math.round(error/beta); 
				}
			}							
		return LSS / (h*w/Math.pow(blocksize,2));	
	}
	
	
	
	public static double calculateESS (String file_name, BlkImgDataSrc new_src, double threshold, int blocksize)
	{
		return calculateESS(file_name, new_src, threshold, blocksize, null);
	}
	
	public static double calculateESS (String file_name, BlkImgDataSrc new_src,  double threshold, int blocksize, ImgPane pane)
	{
		double ESS = 0;
		try
		{
			ImgReaderPGM reader = new ImgReaderPGM(file_name);
			ESS = calculateESS((BlkImgDataSrc) reader, new_src, threshold, blocksize, pane); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}			
		return ESS; 
	}
	
	
	/** 
	 * Returns the Edge Similarity Score (ESS) as described in Mao04a, with reasonable parameters
	 * 
	 * @param org_file_name original image file
	 * @param new_file_name reconstructed image file
	 * @param pane optional ImgPane for vizualization, set null to disable
	 * @return ESS 
	 */
	public static double calculateESS (String org_file_name, String new_file_name, ImgPane pane)
	{
		double ESS = 0;
		double threshold = 12;
		int blocksize = 8;
		
		try
		{
			ImgReaderPGM old_reader = new ImgReaderPGM(org_file_name);
			ImgReaderPGM new_reader = new ImgReaderPGM(new_file_name);
			ESS = calculateESS((BlkImgDataSrc) old_reader, (BlkImgDataSrc) new_reader, threshold, blocksize, pane); 
		}
		catch (IOException e)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, e.getMessage());
		}			
		return ESS; 
	}


		
	/**
	 *
	 * Returns the Edge Similarity Score (ESS) as described in Mao04a
	 *  
	 * @param org_src original image
	 * @param new_src reconstructed image
	 * @param threshold threshold for edge detection
	 * @param pane Pass a JVT Image pane for visualization
	 * 	 
	 * @return
	 */
	public static double calculateESS (BlkImgDataSrc org_src, BlkImgDataSrc new_src, double threshold, int blocksize, ImgPane pane)
	{
		DataBlkInt org_data_blk = getDataBlkInt(org_src, 0);
		DataBlkInt new_data_blk = getDataBlkInt(new_src, 0);		
		
		// Visualization
        Image img = null;
        Image img_r = null;
        Image img_diff = null;
        Graphics g = null;
        Graphics g_diff = null;
        Graphics g_r = null;
		if (pane != null)
		{
			img = new BufferedImage(512, 512, BufferedImage.TYPE_INT_RGB);		
			img_diff = new BufferedImage(512, 512, BufferedImage.TYPE_INT_RGB);
			img_r = new BufferedImage(512, 512, BufferedImage.TYPE_INT_RGB);			
			g = img.getGraphics();
			g_diff = img_diff.getGraphics();
			g_r = img_r.getGraphics();
			Image org_img = BlkImgDataSrcImageProducer.createImage(org_src);
			g.drawImage(org_img, 0, 0, Color.RED, null);			
			g_diff.drawImage(pane.getImage(0), 0, 0, Color.RED, null);
			g_r.drawImage(pane.getImage(0), 0, 0, Color.RED, null);
			pane.removeAll();			
	        pane.addImage (img, " Edge Blocks and Non-Edge Blocks (Original)");
	        pane.addImage (img_r, " Edge Blocks and Non-Edge Blocks (Reconstructed)");
	        pane.addImage (img_diff, "Differences");
		}

		
		if ( org_data_blk.w != new_data_blk.w ||  
			 org_data_blk.h != new_data_blk.h)
		{
			//throw (new IllegalArgumentException("Images are not of the same dimension"));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "LSS/ESS: Images are not of the same dimension. ESS not calculated.");
			return 0;
		}
	 				
		int ulx = org_data_blk.ulx;
		int uly = org_data_blk.uly;
		int w = org_data_blk.w;
		int h = org_data_blk.h;
		
		if (w % 8 != 0 || h % 8 != 0)
		{
			//throw (new IllegalArgumentException("Image dimension not dividable by 8. (Not implemented)"));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "LSS/ESS: Image dimension cannot be divided by blocksize. (Not implemented). ESS not calculated.");
			return 0;
		}						  
		
		int i, j ;
		double ESS = 0.0d;
				
		int nd = 0;
		int od = 0;
		Sobel sobel = new Sobel();
		
		double sumW = 0d;
		int sumC = 0;

		//for each block get the representative angle in original and reconstructed img
		for (j = 0; j < h; j+=blocksize)
			for(i = 0; i < w; i+=blocksize)
			{
				DataBlkInt org_blk = new DataBlkInt(0,0,blocksize,blocksize);
				DataBlkInt new_blk = new DataBlkInt(0,0,blocksize,blocksize);
				//get data blocks
			    for(int jj=0; jj<blocksize; jj++)
			    	for (int ii=0; ii<blocksize; ii++)
			    	{
						od = org_data_blk.data[(j+jj)*w + i+ii] + 128; //add DC_OFFSET
						nd = new_data_blk.data[(j+jj)*w + i+ii] + 128; //add DC_OFFSET
						
						nd = (nd < 0) ? 0 : (nd > 255) ? 255 : nd; //check range						
						
						org_blk.data[jj*blocksize + ii] = od; 
						new_blk.data[jj*blocksize + ii] = nd;												
			    	}
			  
			   
			   int angleID_org = sobel.imageReprAngleID(org_blk, threshold);
			   int angleID_new = sobel.imageReprAngleID(new_blk, threshold);
			   
			   if (angleID_new != 0 || angleID_org != 0)
			   {
			   		sumC++;
			   }
			   if (angleID_new != 0 && angleID_org != 0)
			   {		
			   		/*if (angleID_new != angleID_org)
			   		{
			   			System.out.println ("Different !" + Math.abs(Math.cos(sobel.IdtoReprAngl(angleID_org) - sobel.IdtoReprAngl(angleID_new))));
			   		}*/
			   		sumW += Math.abs(Math.cos(sobel.IdtoReprAngl(angleID_org) - sobel.IdtoReprAngl(angleID_new)));
			   }
			   
			   if (g != null)
			   {
			   	if (angleID_org != 0)
			   	{
			   		g.setColor(Color.red);
			   		g_diff.setColor(Color.red);
			   	}
			   	else
			   	{
			   		g.setColor(Color.blue);
			   		g_diff.setColor(Color.blue);
			   	}
				g.drawRect(i+2, j+2, blocksize-2, blocksize-2);

				if (angleID_new != angleID_org)
			   	{
			   		g_diff.drawRect(i+3, j+3, blocksize-3, blocksize-3);			   		
			   	}
			   	
			   	
			   	if (angleID_new != 0)
			   	{
			   		g_r.setColor(Color.red);
			   		g_diff.setColor(Color.red);
			   	}
			   	else
			   	{
			   		g_r.setColor(Color.blue);
			   		g_diff.setColor(Color.blue);
			   	}
				if (angleID_new != angleID_org)
			   	{
			   		g_diff.drawRect(i+4, j+4, blocksize-5, blocksize-5);			   		
			   	}
				g_r.drawRect(i+2, j+2, blocksize-2, blocksize-2);

			   }			    
			}		
		//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "sumW=" + sumW + " sumC=" + sumC);
		
		//if the image is very smooth, i.e. there are no egde blocks, assign an ESS of 0.5
		if (sumC == 0)
		{
			ESS = 0.5;
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "No edge blocks found. assigning ESS of 0.5");
		}
		else
		{
			ESS = sumW / (double) sumC;
		}	        			        		
		return ESS;	

	}
	
	public static void writePGM (BlkImgDataSrc src, String fname, int c)
	{
		writePGM(getDataBlkInt(src,c), fname);
	}
	
	/**
	 * Write data to PGM, default offset of 128
	 * @param data_blk
	 * @param fname
	 * @param offset
	 */
	public static void  writePGM (DataBlkInt data_blk, String fname)
	{
		writePGM (data_blk, fname, 128);
	}
	
	/**
	 * 
	 * Write Data to PGM
	 * 
	 * NOTE that this function does not regard components or tilings!
	 * 
	 * Write Data to PGM FILE
	 * 
	 * @param src a BlkImgDataSrc providing the data
	 */
	
	public static void  writePGM (DataBlkInt data_blk, String fname, int offset)
	{
		int ulx = data_blk.ulx;
		int uly = data_blk.uly;
		int w = data_blk.w;
		int h = data_blk.h;
		
		int data  [] = data_blk.data;						
		int i, j , val;

		
		try {
			//create a buffered output stream
			FileOutputStream fos = new FileOutputStream (fname);
			BufferedOutputStream bout = new BufferedOutputStream (fos);
			DataOutputStream out = new DataOutputStream(bout);
			
			//Write PGM Header
			out.writeBytes("P2\n" + w + " " + h + "\n255\n");
			
			
			//write data, shifting values back to original range 0..255
			for (j = 0; j < h; j++)
			{
				for(i = 0; i < w; i++)
				{

					val = data[j*w + i];										
					//val = val & 0xFF;
					
					//Shift back to original range
					val = val + offset; 
					val = (val < 0) ? 0 : (val > 255) ? 255 : val; //check range
					/*if ((val < 0) || (val > 255))
					{
						FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "val = " + val);
					}*/
					out.writeBytes (Integer.toString( val) + " " );
				}
				out.writeBytes("\n");
			}
			out.flush();
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
		finally
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Written to " + fname);
		}
	}
	
	/**
	 * 
	 * Write Data to PBM
	 * 
	 * NOTE that this function does not regard components or tilings!
	 * 
	 * Write Data to PBM FILE
	 * 
	 * @param src a BlkImgDataSrc providing the data
	 */
	
	public static void  writePBM (DataBlkInt data_blk, String fname)
	{
		int w = data_blk.w;
		int h = data_blk.h;
		
		int data  [] = data_blk.data;						
		int i, j , val;

		
		try {
			//create a buffered output stream
			FileOutputStream fos = new FileOutputStream (fname);
			BufferedOutputStream bout = new BufferedOutputStream (fos);
			DataOutputStream out = new DataOutputStream(bout);
			
			//Write PBM Header
			out.writeBytes("P1\n" + w + " " + h + "\n");
			
			
			//write data, shifting values back to original range 0..255
			for (j = 0; j < h; j++)
			{
				for(i = 0; i < w; i++)
				{
					val = data[j*w + i];										
					out.writeBytes (Integer.toString( val) + " " );
				}
				out.writeBytes("\n");
			}
			out.flush();
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
		finally
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Written to " + fname);
		}
	}
	
	
	
	/**
	 * Write a single string value to a result file 
	 * 
	 * @param fname filename of the gnuplot file
	 * @param res result value 
	 * @param lf include linefeed?
	 */
	public static void GPwriteResult(String fname, String res, boolean lf)
	{
		try
		{
			
			FileOutputStream fos = new FileOutputStream (fname, true);
			DataOutputStream out = new DataOutputStream(fos);
			//if (res < 0) res = 0;
			//if (Double.isNaN(res)) res = 0;
			out.writeBytes(res + " " + (lf ? "\n" : ""));		
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
		finally
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "GP appended to " + fname);	
		
		}
		
	}
	
	
	/**
	 * Write a single value to a result file 
	 * 
	 * @param fname filename of the gnuplot file
	 * @param res result value 
	 * @param lf include linefeed?
	 */
	public static void GPwriteResult(String fname, double res, boolean lf)
	{
		try
		{
			
			FileOutputStream fos = new FileOutputStream (fname, true);
			DataOutputStream out = new DataOutputStream(fos);
			//if (res < 0) res = 0;
			//if (Double.isNaN(res)) res = 0;
			out.writeBytes(Double.toString(res) + " " + (lf ? "\n" : ""));		
			out.close();
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "File not Found");
		}
		catch (IOException ioe) {			
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "IO Error");	
		}
		finally
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "GP appended to " + fname);	
		
		}
		
	}

}

	

