/*
 * Created on Jun 13, 2006
 *
 */
package jj2000.j2k.wavelet;

import java.util.Iterator;
import java.util.Vector;

import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

public class QTDM {


	public final static int FUNC_LIN = 0;
	public final static int FUNC_RES = 1;
	public final static int FUNC_LEV = 2;
	
	
	public static double compareCommonLeaves (Subband tree1, Subband tree2, int max_rl, String sfunc)
	{		
		Vector<Vector<Subband>> leaves1 = new Vector<Vector<Subband>>(max_rl+1);
		Vector<Vector<Subband>> leaves2 = new Vector<Vector<Subband>>(max_rl+1);
		Vector<Subband> uniqueLeaves1 = new Vector<Subband>();
		Vector<Subband> uniqueLeaves2 = new Vector<Subband>();
		Vector<Subband> uniqueLeaves = new Vector<Subband>();		
		Vector<Subband> allLeaves = new Vector<Subband>();
		int scoreUnique;
		int scoreAll;
		double distance; 
		
		for (int l = 0; l <= max_rl; l ++)
		{
			leaves1.add(new Vector<Subband>());
			leaves2.add(new Vector<Subband>());
		}
		getLeavesRecursive(tree1, leaves1);
		getLeavesRecursive(tree2, leaves2);
		for (int l = 0; l <= max_rl; l ++)
		{
			allLeaves.addAll(leaves1.get(l));
			allLeaves.addAll(leaves2.get(l));			
			uniqueLeaves1.addAll(leaves1.get(l));
			uniqueLeaves2.addAll(leaves2.get(l));
		}
		removeNonUniqueLeaves(uniqueLeaves1, uniqueLeaves2);
		uniqueLeaves.addAll(uniqueLeaves1);
		uniqueLeaves.addAll(uniqueLeaves2);
		
		int func = 0;
		if (sfunc.compareToIgnoreCase("lin") == 0)
		{
			func = FUNC_LIN;
		} else if (sfunc.compareToIgnoreCase("res") == 0) {
			func = FUNC_RES;
		} else if (sfunc.compareToIgnoreCase("lev") == 0) {
			func = FUNC_LEV;
		} else	{
			System.err.println ("Could not parse evaluation function from \"" + sfunc + "\"");
			System.exit(1);
		}
				
		scoreAll = evaluate(allLeaves, max_rl, func);
		scoreUnique = evaluate(uniqueLeaves, max_rl, func);
		distance = (double) scoreUnique / (double) scoreAll;
		
		return distance;
	}
	
	private static void getLeavesRecursive (Subband sb, Vector<Vector<Subband>> leaves)
	{
		if (!sb.isNode)
		{
			leaves.get(sb.resLvl).add(sb);
			//leaves[sb.resLvl].add(sb);
		} else {
			getLeavesRecursive(sb.getLL(), leaves);
			getLeavesRecursive(sb.getLH(), leaves);
			getLeavesRecursive(sb.getHL(), leaves);
			getLeavesRecursive(sb.getHH(), leaves);
		}		
	}
	
	private static void removeNonUniqueLeaves (Vector<Subband> leaves1, Vector<Subband> leaves2)
	{
		Iterator<Subband> it1 = leaves1.iterator();
		Subband sb1;
		Subband sb2;
	
		while (it1.hasNext())
		{
			sb1 = it1.next();
			Iterator<Subband> it2 = leaves2.iterator();
			//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, " sbb1:" + sb1 );
			while(it2.hasNext())
			{
				sb2 = it2.next();

				//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, " sbb2:\n" + sb2 );
				if (isSameSb(sb1,sb2)) {
					FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Same sbbs: (" + sb1.sbandIdx + "," + sb1.resLvl + ") and (" + sb2.sbandIdx + "," + sb2.resLvl + ")");
					it1.remove();
					it2.remove();
					//FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Same!");
				}
			}			
		}
	}
	
	private static int evaluate (Vector<Subband> uniqueLeaves, int max_l, int func)
	{
		switch (func) {
		case FUNC_LEV:
			return evaluate_lev(uniqueLeaves, max_l);
		case FUNC_LIN:
			return evaluate_lin(uniqueLeaves, max_l);
		case FUNC_RES:
			return evaluate_res(uniqueLeaves, max_l);
		default:
			System.err.println("Unrecognized evaluation function code " + func);
			System.exit(1);
			return 0;
		}
		
	}
	
	private static int evaluate_lev(Vector<Subband> uniqueLeaves, int max_l)
	{
		Iterator<Subband> it = uniqueLeaves.iterator();
		int score = 0;
		
		while (it.hasNext()) {
			Subband sb = (Subband) it.next();
			//high score for small level --> 
			//high score for large subbands
			score += Math.pow(2, (max_l - sb.level));
		}
		
		return score;		
	}
	
	private static int evaluate_lin(Vector<Subband> uniqueLeaves, int max_l)
	{
		return uniqueLeaves.size();
	}

	private static int evaluate_res(Vector<Subband> uniqueLeaves, int max_l)
	{
		Iterator<Subband> it = uniqueLeaves.iterator();
		int score = 0;
		
		while (it.hasNext()) {
			Subband sb = (Subband) it.next();
			//high score for low resolution
			score += Math.pow(2, (max_l - sb.resLvl));
		}
		
		return score;		
	}
	
	private static boolean isSameSb (Subband sb1, Subband sb2)
	{
		if (//sb1.orientation == sb2.orientation && //dom: removed orientation  
				sb1.ulx == sb2.ulx && 
				sb1.uly == sb2.uly && 
				sb1.w == sb2.w && 
				sb1.h == sb2.h) {
			return true;
		} else
			return false;
	}

}
