/*
 * Created on Mar 8, 2005
 *
 */
package jj2000.j2k.wavelet;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;

import jj2000.j2k.image.BlkImgDataSrc;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

/**
 * 
 * This class generates a pseudo-random structure based on a uniform distribution of all 
 * subband structures
 * 
 * @author dengel
 *
 */
public class WPUniformGenerator {

	public static int SCALE = 30; //precision for probabilities
	
	/*public int min_decomp_detail;
	public int max_decomp_detail;
	public int max_decomp_approx;
	public int min_decomp_approx;
	public float base_value;
	public float change_factor;
	public float min_squareness_approx;
	public float min_squareness_detail;*/
	public int max_decomp_global;
	//public int pyramidal_level = 0;
	public long seed;	
	
	private BlkImgDataSrc src;
	
	private Random randomGen;
	
	public Vector pattern = new Vector();
	

	/**
	 * create random structure generator without imgsrc 
	 * this will only create a subband tree, but the subbands 
	 * will not have a height or a width
	 * 
	 */
	public WPUniformGenerator(String spec) {		
		parseSpec(spec);		
	}

	
	/**
	 * 
	 * create random structure generator with imgsrc
	 * use this to create a subband tree and give each 
	 * subband its approriate height and widht, 
	 * based on imgsrc
	 * 
	 */
	public WPUniformGenerator(String spec, BlkImgDataSrc imgsrc) {
		this.src = imgsrc;		
		parseSpec(spec);		
	}
		
	/**
	 * create a uniformly distributed (pseudo)random wavelet packet decomposition structure
	 * @return
	 */
	public SubbandStructure createRandomStructure()
	{
		SubbandStructure sbbs;
		if (src != null) 
			sbbs = new SubbandStructure(src.getImgWidth(), src.getImgHeight());
		else
			sbbs = new SubbandStructure();

		randomGen = new Random (seed);
		
		createRandomStructureRecursive(sbbs, max_decomp_global);
		
		Subband sb = sbbs.getDeepestGlobalLL();
		if (sb.level != sbbs.resLvl)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Random WPS set new resolution level to: " + sb.level);
			sbbs.setResolutionLevelRecursive(sb.level);
		}
		return sbbs;
	}
	
	
	private void createRandomStructureRecursive(Subband sb, int j)
	{		
		
		if (j==0) {
//			System.out.println (" j = 0; we're done");
			return;
		}
		
		BigDecimal r = new BigDecimal(randomGen.nextDouble());
		BigDecimal num = new BigDecimal (A(j));
//		System.out.println (" Level " + j + ": " + num + " decompositions" );
		
		BigDecimal prob = BigDecimal.ONE.divide(num, SCALE, BigDecimal.ROUND_HALF_EVEN);		
//		System.out.println (" prob =  " + prob);
//		System.out.println (" r =  " + r);
		
		if (r.compareTo(prob) > 0) {
			sb.split(null, null);
			createRandomStructureRecursive(sb.getLL(), j-1);
			createRandomStructureRecursive(sb.getHL(), j-1);
			createRandomStructureRecursive(sb.getLH(), j-1);
			createRandomStructureRecursive(sb.getHH(), j-1);
		}
	}
	
	
	/**
	 *  number of trees for level j
	 */
	public static BigInteger A (int j)
	{
		if (j == 0)
			return BigInteger.ONE;
		else {
			return A(j-1).pow(4).add(BigInteger.ONE);
		}
	}

	/**
	 * parse the specifications for this class
	 * 
	 * @param spec specifications as string
	 */
	private void parseSpec(String spec)
	{
		StringTokenizer t = new StringTokenizer(spec, "sg", true);
		
		while (t.hasMoreTokens())
		{
			String token = t.nextToken();
			char var = token.charAt(0);
			switch (var) {
			case 's':
				String sseed = t.nextToken();
				if (sseed.equalsIgnoreCase("t"))
					seed = System.currentTimeMillis();
				else
					seed = Long.parseLong(sseed);								
				break;
			case 'g':
				max_decomp_global = Integer.parseInt(t.nextToken());
				break;			
			default:
				throw new IllegalArgumentException("Malformed specification string for WwpsRandom: " + spec);
			}
		}
		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Random Wavelet Structure Parameters: " + this.toString() );
	}
	
	

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "[WPUniformGenerator: Seed=" + seed + " MaxDecompositionDepthGlobal="
				+ max_decomp_global + "]";
	}
}
