/* 
 * CVS identifier:
 * 
 * $Id: ForwWTFullBBA.java,v 1.8 2008-05-05 11:33:34 dengel Exp $
 * 
 * Class:                   ForwWTFullBBA
 * 
 * Description:             
 * 
 * Created on Feb 3, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */

package jj2000.j2k.wavelet.analysis;

import java.io.IOException;

import jj2000.j2k.encoder.EncoderSpecs;
import jj2000.j2k.image.BlkImgDataSrc;
import jj2000.j2k.image.DataBlk;
import jj2000.j2k.image.DataBlkFloat;
import jj2000.j2k.image.DataBlkInt;
import jj2000.j2k.util.ArrayUtil;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.wavelet.Subband;
import jj2000.j2k.wavelet.WaveletFilter;
import jj2000.j2k.wavelet.analysis.bestbasis.AddCostFunction;

/**
 *
 * The BBA version of ForwWTFull
 * 
 * the order of things happening had to be rearranged as for the BBA the initialization
 * of the subband structure and the decomposition are closely coupled 
 */
public class ForwWTFullBBA extends ForwWTFull {

	
	/** specifies if the global LL subband tree should always be decomposed (regardless of the cost) */ 
	private boolean LLAlwaysDecompose;
	
	/**
	 * The Costfunction to be used for decomposition
	 */
	private AddCostFunction cf;

	/**
	 * @param src
	 * @param encSpec
	 * @param pox
	 * @param poy
	 */
	public ForwWTFullBBA(BlkImgDataSrc src, EncoderSpecs encSpec, int pox,
			int poy) {
		super(src, encSpec, pox, poy);
				
		cf = encSpec.acf.getCostFunction(0,0);
		LLAlwaysDecompose = encSpec.bbaLLAlwaysDecompose;
	}

	/**
	 * create the wp tree for component c
	 * 
	 * also does the decomposition 
	 * 
	 * @param c
	 */
	private void createWPTree(int c) {
		 int k,w,h;
         DataBlk bufblk;
         Object dst_data;
	    
         w = getCompWidth(c);
         h = getCompHeight(c);
         
         intData = (filters.getWTDataType(tIdx,c)==DataBlk.TYPE_INT);
	    
         //Get the source image data
         if(intData){
             decomposedComps[c] = new DataBlkInt(0,0,w,h);
             bufblk = new DataBlkInt();
         }
         else {
             decomposedComps[c] = new DataBlkFloat(0,0,w,h);
             bufblk = new DataBlkFloat();
         }
	    
         // Get data from source line by line (this diminishes the memory
         // requirements on the data source)
         dst_data = decomposedComps[c].getData();
         bufblk.ulx = 0;
         bufblk.w = w;
         bufblk.h = 1;
         for (k=0; k<h; k++) {
             bufblk.uly = k;
             bufblk.ulx = 0;
             bufblk = src.getInternCompData(bufblk,c);
             System.arraycopy(bufblk.getData(),bufblk.offset,
                              dst_data,k*w,w);
         }
                     
         //Decompose source image
         subbTrees[tIdx][c] = new SubbandAn(getCompWidth(c), getCompHeight(c),
          			getULX(c), getULY(c), getDecompLevels(tIdx, c));
         
         waveletTreeDecomposition(decomposedComps[c],subbTrees[tIdx][c],c); 
         
         subbTrees[tIdx][c].calcL2Norms();
	    
         // Make the first subband the current one
         currentSubband[c] = getNextSubband(c);
         lastn[c] = -1;
         lastm[c] = 0;
	}
	

	public SubbandAn getSubbandTree(int t, int c) {	
	     if (subbTrees[t][c] == null) {
			createWPTree(c);
        }
        return subbTrees[t][c];	
        
	}

			
	public void nextTile() {
		imgdatasrc.nextTile();
		tIdx = getTileIdx();
		if (tIdx != 0)
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "nextTile: only one tile supported, you specified more than one ");
	}
	
	public void setTile(int x, int y) {
		imgdatasrc.setTile(x,y);
		tIdx = getTileIdx();
		if (tIdx != 0)
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "setTile: only one tile supported, you specified more than one ");
	}
	
	protected void waveletTreeDecomposition(DataBlk band, SubbandAn subband,
			int c) {		
		
		waveletTreeDecompositionBBA(band, subband, c, subband.resLvl);
		
		// uncomment the next line to see wrong numbering: 
		//subband.plotTreeStructure("tree.eps", 512, 512);
		
		// now we have a WP decomposition, but if the deepest LL subband tree was
		// pruned in the process, the resolution level is set to a wrong value and
		// has to be corrected
		Subband sb = subband.getDeepestGlobalLL();
		if (sb.level != subband.resLvl)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "BBA set new resolution level to: " + sb.level);
		}
		subband.setResolutionLevelRecursive(sb.level);		
		
		// update the new resolutionLevel in the encSpecs, 
		// so the header encoders know about it, too
		dls.setDefault(new Integer(sb.level));
		
		
		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "BBA tree structure: 0x" + (ArrayUtil.printBitArrayAsHex(subband.toBitPattern())));
		
		subband.plotTreeStructure("tree.eps", 512, 512);
	    
		//		 if requested, write structure to file as a hex string
		if ( pl.containsValue("Wwritewpsfile") && pl.getParameter("Wwritewpsfile") != "")
		{
			String fname =  pl.getParameter("Wwritewpsfile");
			try {
				ArrayUtil.writeBitArrayToHexFile(pl.getParameter("Wwritewpsfile"), subband.toBitPattern()); }
			catch (IOException ioe) {
				FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "Could not write WPS to " + fname);
			}
			catch (NullPointerException ne)
			{			
					FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR, "Null pointer exception when trying to write WPS Structure. Possibly you attempted to write it too soon.");			
			}
			finally	{
				FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Wavelet Packet Decomposition Structure written as hex to " + fname);				
			}				
		}			
		
	}
	
	/**
	 * implements the bba recursively 
	 * (with a high demand on memory)
	 * 
	 * @param band
	 * @param subband
	 * @param c
	 * @param lvl
	 * @return
	 */
	private double waveletTreeDecompositionBBA (DataBlk band, SubbandAn subband,
			int c, int lvl) {
		
		double cost = cf.getCost( band, (Subband) subband);
		
		//lowest decomposition level reached
		if (lvl == 0)
		{
			return cost;
		}
		
		//choose filters
		int hi;
		int vi;
		WaveletFilter vfilters[] = filters.getHFilters(tIdx, c);
		WaveletFilter hfilters[] = filters.getVFilters(tIdx, c);
		
		hi = (subband.resLvl <= hfilters.length) ? subband.resLvl - 1 : hfilters.length - 1;
		vi = (subband.resLvl <= vfilters.length) ? subband.resLvl - 1 : vfilters.length - 1;
		
		//save the original coefficients
		
		float[] savedCoeffs_float  = null;
		int[] savedCoeffs_int = null;
	
		if (intData) {
			savedCoeffs_int = new int[band.h*band.w];
			System.arraycopy( ((DataBlkInt)band).getDataInt(), 0, savedCoeffs_int, 0, savedCoeffs_int.length);			
			
		} else {
			savedCoeffs_float = new float[band.h*band.w];
			System.arraycopy( ((DataBlkFloat)band).getDataFloat(), 0, savedCoeffs_float, 0, savedCoeffs_float.length);			
		}
		
		//split subband, set filters
		subband.split(hfilters[hi], vfilters[vi]);  
				
        //Perform the 2D wavelet decomposition of the current subband
        wavelet2DDecomposition(band, (SubbandAn)subband, c);
        
        //Go on with children
        double childcost = 0;
        
        
        childcost += waveletTreeDecompositionBBA(band, (SubbandAn)subband.getLL(), c, lvl-1);
        childcost += waveletTreeDecompositionBBA(band, (SubbandAn)subband.getHL(), c, lvl-1);
        childcost += waveletTreeDecompositionBBA(band, (SubbandAn)subband.getLH(), c, lvl-1);
        childcost += waveletTreeDecompositionBBA(band, (SubbandAn)subband.getHH(), c, lvl-1);        
        
		//System.out.println ("rl " + subband.resLvl + " Cost=" + cost + " Childcost=" + childcost);
        
        if (cost <= childcost	 //restore if childcost is more than the parent cost 
        		&& subband.level != 0 //but always decompose the root
					&& 
					! (LLAlwaysDecompose &&
					subband.gOrient == Subband.WT_ORIENT_LL) //and also always  
					//decompose the global LL subband if ensureLLOnDeepestRes is set												
				)
        {        	
        	if (intData) {
        		band.setData(savedCoeffs_int);
        	} else {
        		band.setData(savedCoeffs_float);
        	}
        	subband.bereave();      
        }
        else
        	cost = childcost;
        
        return cost;
	}
	
}
