/* 
 * CVS identifier:
 * 
 * $Id: AddCostFunctionSpec.java,v 1.3 2008-05-05 11:33:34 dengel Exp $
 * 
 * Class:                   AddCostFunctionSpec
 * 
 * Description:             
 * 
 * Created on Feb 4, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */

package jj2000.j2k.wavelet.analysis.bestbasis;

import jj2000.j2k.ModuleSpec;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.util.ParameterList;

/**
 * 
 * specifies the costfunction to be used in conjunction with BBA
 *
 */
public class AddCostFunctionSpec extends ModuleSpec {
	
	private AddCostFunction defaultCostFunction;

	/**
	 *  Constructs AddCostFunctionSpec
	 * 
	 * @param nt
	 * @param nc
	 * @param type
	 */
	public AddCostFunctionSpec(int nt, int nc, byte type, ParameterList pl) {
		super(nt, nc, type);
		
		if (! pl.containsKey("Wbbacf"))
		{
			defaultCostFunction = null;
			return;
		}
		
		String cfs = pl.getParameter("Wbbacf");
				
		
		if (cfs.equalsIgnoreCase("LogE"))
		{
			defaultCostFunction = new LogEnergy();
		}
		else if (cfs.equalsIgnoreCase("EIC"))
		{
			defaultCostFunction = new EntropyInformationCost();
		}
		else if (cfs.equalsIgnoreCase("Norml"))
		{
			defaultCostFunction = new Norml();
		}
		else if (cfs.equalsIgnoreCase("Norml2"))
		{
			defaultCostFunction = new Norml2();
		}		
		else
		{			 
			throw new IllegalArgumentException("Unknown costfunction '" + cfs + "'");
		}
		setDefault(defaultCostFunction);
	}
	
	
	public AddCostFunction getCostFunction(int t, int c)
	{
		return defaultCostFunction;
	}

}
